/*
 * aepsilon.cpp
 *
 *  Created on: 24. 2. 2014
 *      Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>

#include "automaton/simplify/EpsilonRemoverIncoming.h"
#include "automaton/simplify/EpsilonRemoverOutgoing.h"
#include "automaton/simplify/efficient/EpsilonRemoverIncoming.h"
#include "automaton/simplify/efficient/EpsilonRemoverOutgoing.h"
#include "grammar/simplify/EpsilonRemover.h"

int main(int argc, char** argv) {

	try {
		TCLAP::CmdLine cmd("Automaton rename binary", ' ', "0.01");

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Automaton/Grammar where to remove epsilon transitions",		false,	"-",		"file");
		cmd.add( input );

		std::vector<std::string> algorithms{"incoming", "outgoing", "default" };
		TCLAP::ValuesConstraint<std::string> allowedAlgorithms( algorithms );
		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",		"Specifies algorithm to use",		false,	"default",	&allowedAlgorithms);
		cmd.add( algorithm );

		TCLAP::SwitchArg efficient(		"e",	"efficient",		"Use efficient implementation",		false);
		cmd.add( efficient );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		std::chrono::measurements::start("Overal", std::chrono::measurements::Type::OVERALL);
		std::chrono::measurements::start("Input read", std::chrono::measurements::Type::AUXILARY);

		std::deque<sax::Token> tokens;
		if(input.isSet()) {
			if(input.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(tokens);
			} else {
				sax::SaxParseInterface::parseFile(input.getValue(), tokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(tokens);
		}

		if(algorithm.getValue() == "outgoing") {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			if(efficient.getValue()) {
				automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverOutgoing::remove(automaton);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

				alib::XmlDataFactory::toStdout(res);
			} else {
				automaton::Automaton res = automaton::simplify::EpsilonRemoverOutgoing::remove(automaton);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

				alib::XmlDataFactory::toStdout(res);
			}
		} else if(algorithm.getValue() == "incoming") {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			if(efficient.getValue()) {
				automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverIncoming::remove(automaton);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

				alib::XmlDataFactory::toStdout(res);
			} else {
				automaton::Automaton res = automaton::simplify::EpsilonRemoverIncoming::remove(automaton);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

				alib::XmlDataFactory::toStdout(res);
			}
		} else if(algorithm.getValue() == "default") {
			if(alib::XmlDataFactory::first<automaton::Automaton>(tokens)) {
				automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

				if(efficient.getValue()) {
					automaton::Automaton res = automaton::simplify::EpsilonRemoverIncoming::remove(automaton);

					std::chrono::measurements::end();
					std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

					alib::XmlDataFactory::toStdout(res);
				} else {
					automaton::Automaton res = automaton::simplify::efficient::EpsilonRemoverIncoming::remove(automaton);

					std::chrono::measurements::end();
					std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

					alib::XmlDataFactory::toStdout(res);
				}
			} else if(alib::XmlDataFactory::first<grammar::Grammar>(tokens)) {
				grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(tokens);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

				grammar::Grammar res = grammar::simplify::EpsilonRemover::remove(grammar);

				std::chrono::measurements::end();
				std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

				alib::XmlDataFactory::toStdout(res);
			} else
				throw exception::AlibException("Invalid formalism");
		} else {
			throw exception::AlibException("Invalid algorithm");
		}

		std::chrono::measurements::end();
		std::chrono::measurements::end();

		if(measure.getValue()) std::clog << std::chrono::measurements::results() << std::endl;

		return 0;
	} catch (const exception::AlibException& exception) {
		alib::XmlDataFactory::toStdout(exception);
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
