/*
 * EpsilonClosure.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "EpsilonClosure.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

#include "../../regexp/properties/RegExpEpsilon.h"

namespace automaton {

namespace properties {

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::EpsilonNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		auto tos = fsm.getTransitions( ).find(std::make_pair(p, std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON) ) );
		if(tos == fsm.getTransitions().end()) continue;

		for( const auto & to : tos->second )
			if( visited [ to ] == false )
				queue.push( to );
	}

	return closure;
}

auto EpsilonClosureEpsilonNFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::EpsilonNFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::MultiInitialStateNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureMultiInitialStateNFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::MultiInitialStateNFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::NFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureNFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::NFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::DFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

auto EpsilonClosureDFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::DFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::ExtendedNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

auto EpsilonClosureExtendedNFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::ExtendedNFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::CompactNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( transition.first.second.getContent().size() == 0 )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

auto EpsilonClosureCompactNFA = EpsilonClosure::RegistratorWrapper<std::set<automaton::State>, automaton::CompactNFA>(EpsilonClosure::getInstance(), EpsilonClosure::epsilonClosure);

std::set<automaton::State> EpsilonClosure::epsilonClosure(const Automaton& automaton, const automaton::State& q) {
	return getInstance().dispatch(automaton.getData(), q);
}

} /* namespace properties */

} /* namespace automaton */

