/*
 * Accept.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Accept.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/TA/NFTA.h>

#include <deque>

namespace automaton {

namespace run {

bool Accept::accept(const automaton::Automaton& automaton, const alib::Object& object) {
	return getInstance().dispatch(automaton.getData(), object.getData());
}

bool Accept::accept(const automaton::Automaton& automaton, const string::LinearString& string) {
	return getInstance().dispatch(automaton.getData(), string);
}

bool Accept::accept(const automaton::Automaton& automaton, const tree::RankedTree& tree) {
	return getInstance().dispatch(automaton.getData(), tree);
}

bool Accept::accept(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) return false;
		state = next->second;
	}
	return automaton.getFinalStates().count(state);
}

auto AcceptDFALinearString = Accept::RegistratorWrapper<bool, automaton::DFA, string::LinearString>(Accept::getInstance(), Accept::accept);

State Accept::calculateState(const automaton::DFTA& automaton, const tree::RankedNode & node) {
	std::vector<State> states;
	states.reserve(node.getSymbol().getRank().getData());
	for (const auto & child : node.getChildren()) {
		State state = calculateState(automaton, *child);
		if (state == FAILSTATE) return FAILSTATE;
		states.push_back(state);
	}
	const auto & it = automaton.getTransitions().find(std::make_pair(node.getSymbol(), states));
	if (it == automaton.getTransitions().end()) return FAILSTATE;
	return it -> second;
}

bool Accept::accept(const automaton::DFTA& automaton, const tree::RankedTree& tree) {
	return automaton.getFinalStates().count(calculateState(automaton, tree.getRoot()));
}

auto AcceptDFTARankedTree = Accept::RegistratorWrapper<bool, automaton::DFTA, tree::RankedTree>(Accept::getInstance(), Accept::accept);

std::set<State> calculateStates(const automaton::NFTA& automaton, const tree::RankedNode & node) {
	std::vector<std::set<State> > states;
	states.reserve(node.getSymbol().getRank().getData());
	for (const auto & child : node.getChildren()) {
		std::set<State> childStates = calculateStates(automaton, *child);
		if (childStates.empty()) return childStates;
		states.push_back(childStates);
	}

	std::set<State> res;
	for (const auto & transition : automaton.getTransitions()) {
		if (transition.first.first != node.getSymbol()) continue;
		int i = (int) transition.first.first.getRank().getData() - 1; //TODO az budou testy upravit na unsigned
		for(; i >= 0 ; i--)
			if (!states[i].count(transition.first.second[i])) break;
		
		if (i == -1) res.insert(transition.second.begin(), transition.second.end());
	}
	return res;
}

bool Accept::accept(const automaton::NFTA& automaton, const tree::RankedTree & tree) {
	std::set<State> res = calculateStates(automaton, tree.getRoot());
	for (auto & state : res)
		if (automaton.getFinalStates().count(state)) return true;
	return false;
}

auto AcceptNFTARankedTree = Accept::RegistratorWrapper<bool, automaton::NFTA, tree::RankedTree>(Accept::getInstance(), Accept::accept);

bool recursiveAccept(const automaton::NFA& automaton, const automaton::State& state, const string::LinearString& string) {
	if( string.getContent().empty( ) )
		return automaton.getFinalStates().count( state );

	const alphabet::Symbol & inputSymbol = string.getContent().front( );

	string::LinearString newString(std::vector<alphabet::Symbol>{ string.getContent().begin() + 1, string.getContent().end()});

	for( const auto & transition : automaton.getTransitionsFromState( state ) )
		for(const auto& to : transition.second)
			if( transition.first.second == inputSymbol && recursiveAccept( automaton, to, newString ) )
				return true;

	return false;
}

bool Accept::accept( const automaton::NFA& automaton, const string::LinearString & string ) {
	return recursiveAccept(automaton, automaton.getInitialState(), string);
}

auto AcceptNFALinearString = Accept::RegistratorWrapper<bool, automaton::NFA, string::LinearString>(Accept::getInstance(), Accept::accept);

bool Accept::accept(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		int pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			int popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(int i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return false;
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);
	}
	if(automaton.getFinalStates().empty()) {
		return pushdownStore.empty();
	} else {
		return automaton.getFinalStates().count(state);
	}
}

auto AcceptDPDALinearString = Accept::RegistratorWrapper<bool, automaton::DPDA, string::LinearString>(Accept::getInstance(), Accept::accept);

} /* namespace run */

} /* namespace automaton */
