/*
 * Occurrences.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Occurrences.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/TA/DFTA.h>

#include <deque>

namespace automaton {

namespace run {

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const alib::Object& object) {
	return getInstance().dispatch(automaton.getData(), object.getData());
}

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const string::LinearString& string) {
	return getInstance().dispatch(automaton.getData(), string);
}

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const tree::RankedTree& tree) {
	return getInstance().dispatch(automaton.getData(), tree);
}

std::set<unsigned> Occurrences::occurrences(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	unsigned i = 0;
	std::set<unsigned> occ;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		if(automaton.getFinalStates().count(state))
			occ.insert(i);

		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) { //makes sence to reset the automaton when there is "unknown" symbol on the input
			state = automaton.getInitialState();
		} else {
			state = next->second;
		}

		i++;
	}

	if(automaton.getFinalStates().count(state))
		occ.insert(i);

	return occ;
}

auto OccurrencesDFALinearString = Occurrences::RegistratorWrapper<std::set<unsigned>, automaton::DFA, string::LinearString>(Occurrences::getInstance(), Occurrences::occurrences);

State Occurrences::occurrences(const automaton::DFTA & automaton, const tree::RankedNode & node, std::set<unsigned> & occ, unsigned & i) {
	std::vector<State> states;
	unsigned tmp = i;
	i++;
	states.reserve(node.getSymbol().getRank().getData());
	bool fail = false;
	for (const auto & child : node.getChildren()) {
		State state = occurrences(automaton, *child, occ, i);
		if (state == FAILSTATE) fail = true;
		else if (!fail) states.push_back(state);
	}
	if (fail) return FAILSTATE;
	const auto & it = automaton.getTransitions().find(std::make_pair(node.getSymbol(), states));
	if (it == automaton.getTransitions().end()) return FAILSTATE;
	State state = it -> second;
	if (automaton.getFinalStates().count(state)) occ.insert(tmp);
	return state;
}

std::set<unsigned> Occurrences::occurrences(const automaton::DFTA& automaton, const tree::RankedTree & tree) {
	std::set<unsigned> occ;
	unsigned i = 0;
	occurrences(automaton, tree.getRoot(), occ, i);
	return occ;
}

auto OccurrencesDFTARankedTree = Occurrences::RegistratorWrapper<std::set<unsigned>, automaton::DFTA, tree::RankedTree>(Occurrences::getInstance(), Occurrences::occurrences);

std::set<unsigned> Occurrences::occurrences(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;
	unsigned i = 0;
	std::set<unsigned> occ;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		if(automaton.getFinalStates().count(state))
			occ.insert(i);

		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		unsigned pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			unsigned popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(unsigned i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return {};
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);

		i++;
	}

	if(automaton.getFinalStates().count(state))
		occ.insert(i);

	if(pushdownStore.empty()) {
		return occ;
	} else {
		return {};
	}
}

auto OccurrencesDPDALinearString = Occurrences::RegistratorWrapper<std::set<unsigned>, automaton::DPDA, string::LinearString>(Occurrences::getInstance(), Occurrences::occurrences);

} /* namespace run */

} /* namespace automaton */
