/*
 * Result.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Result.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/TA/DFTA.h>
#include <label/LabelSetLabel.h>

#include <deque>

namespace automaton {

namespace run {

label::Label Result::result(const automaton::Automaton& automaton, const alib::Object& object) {
	return getInstance().dispatch(automaton.getData(), object.getData());
}

label::Label Result::result(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
		state = next->second;
	}
	return state.getName();
}

auto ResultDFALinearString = Result::RegistratorWrapper<label::Label, automaton::DFA, string::LinearString>(Result::getInstance(), Result::result);

label::Label Result::result(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		int pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			int popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(int i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);
	}
	if(automaton.getFinalStates().empty()) {
		return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
	} else {
		return state.getName();
	}
}

auto ResultDPDALinearString = Result::RegistratorWrapper<label::Label, automaton::DPDA, string::LinearString>(Result::getInstance(), Result::result);

automaton::State Result::calculateState(const automaton::DFTA& automaton, const tree::RankedNode & node) {
	std::vector<automaton::State> states;
	states.reserve(node.getSymbol().getRank().getData());
	for (const auto & child : node.getChildren()) {
		automaton::State state = calculateState(automaton, *child);
		if (state == FAILSTATE) return FAILSTATE;
		states.push_back(state);
	}
	const auto & it = automaton.getTransitions().find(std::make_pair(node.getSymbol(), states));
	if (it == automaton.getTransitions().end()) return FAILSTATE;
	return it -> second;
}

label::Label Result::result(const automaton::DFTA& automaton, const tree::RankedTree& tree) {
	State state = calculateState(automaton, tree.getRoot());
	return state.getName();
}

auto ResultDFTARankedTree = Result::RegistratorWrapper<label::Label, automaton::DFTA, tree::RankedTree>(Result::getInstance(), Result::result);

} /* namespace run */

} /* namespace automaton */
