/*
 * ToAutomaton.cpp
 *
 * Created on: 1. 11. 2013
 * Author: Tomas Pecka
 */
#include "ToAutomaton.h"

#include <alphabet/LabeledSymbol.h>
#include <exception/AlibException.h>

namespace grammar {

namespace convert {

automaton::Automaton ToAutomaton::convert(const grammar::Grammar& grammar) {
	 return getInstance().dispatch(grammar.getData());
}

automaton::NFA ToAutomaton::convert(const grammar::LeftRG& grammar) {
	std::map<alphabet::Symbol, automaton::State> stateMap;
	std::set<automaton::State> states;

	// step 2
	for(const auto& symbol : grammar.getNonterminalAlphabet()) {
		automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	const automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	states.insert(q0);
	automaton::NFA automaton(q0);
	automaton.setInputAlphabet(grammar.getTerminalAlphabet());
	automaton.setStates(states);

	// step 3
	for(const auto& rule : grammar.getRules()) {
		const alphabet::Symbol& lhs = rule.first;
		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) { // if B->Ca => \delta(C,a)=B
				const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				automaton.addTransition(stateMap.find(rhs.first)->second, rhs.second, stateMap.find(lhs)->second);
			}
			else { // if B->a => \delta(StartState,a)=B
				const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
				automaton.addTransition(q0, rhs, stateMap.find(lhs)->second);
			}
		}
	}

	// step 5
	automaton.addFinalState(stateMap.find(grammar.getInitialSymbol())->second);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(q0);

	return automaton;
}

auto ToAutomatonLeftRG = ToAutomaton::RegistratorWrapper<automaton::NFA, grammar::LeftRG>(ToAutomaton::getInstance(), ToAutomaton::convert);

automaton::NFA ToAutomaton::convert(const grammar::RightRG& grammar) {
	std::map<alphabet::Symbol, automaton::State> stateMap;
	std::set<automaton::State> states;

	// step2
	for(const auto& symbol : grammar.getNonterminalAlphabet()) {
		automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	const automaton::State AState = automaton::createUniqueState(automaton::State("A"), states);
	states.insert(AState);
	automaton::NFA automaton(stateMap.find(grammar.getInitialSymbol())->second);
	automaton.setStates(states);

	automaton.setInputAlphabet(grammar.getTerminalAlphabet());

	// step 3
	for(const auto& rule : grammar.getRules()) {
		const alphabet::Symbol& lhs = rule.first;
		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) { // if B->aC => \delta(B,a)=C
				const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs.first, stateMap.find(rhs.second)->second);
			}
			else { // if B->a => \delta(B,a)=AState
				const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs, AState);
			}
		}
	}

	// step 5
	automaton.addFinalState(AState);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(automaton.getInitialState());

	return automaton;
}

auto ToAutomatonRightRG = ToAutomaton::RegistratorWrapper<automaton::NFA, grammar::RightRG>(ToAutomaton::getInstance(), ToAutomaton::convert);

template <class T>
automaton::NPDA ToAutomaton::convert(const T& grammar) {
	automaton::NPDA automaton(automaton::State('q'), grammar.getInitialSymbol());

	automaton.setInputAlphabet(grammar.getTerminalAlphabet());

	automaton.setStackAlphabet(grammar.getNonterminalAlphabet());
	for(const alphabet::Symbol& symbol : grammar.getTerminalAlphabet())
		automaton.addStackSymbol(symbol);

	//for(const std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& kv : grammar.getRules())
	for(const auto& kv : grammar.getRules())
		for(const std::vector<alphabet::Symbol>& rhs : kv.second)
			automaton.addTransition(automaton.getInitialState(), std::vector<alphabet::Symbol>{kv.first}, automaton.getInitialState(), rhs);

	for(const alphabet::Symbol& symbol : grammar.getTerminalAlphabet())
		automaton.addTransition(automaton.getInitialState(), symbol, std::vector<alphabet::Symbol>{symbol}, automaton.getInitialState(), std::vector<alphabet::Symbol>{});

	return automaton;
}

auto ToAutomatonCFG = ToAutomaton::RegistratorWrapper<automaton::NPDA, grammar::CFG>(ToAutomaton::getInstance(), ToAutomaton::convert);
auto ToAutomatonEpsilonFreeCFG = ToAutomaton::RegistratorWrapper<automaton::NPDA, grammar::EpsilonFreeCFG>(ToAutomaton::getInstance(), ToAutomaton::convert);

} /* namespace convert */

} /* namespace grammar */
