/*
 * InputDrivenDPDA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "InputDrivenDPDA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>
#include <algorithm>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace automaton {

InputDrivenDPDA::InputDrivenDPDA(State initialState, alphabet::Symbol initialPushdownSymbol) : SingleInitialSymbolPushdownStoreAlphabet(std::move(initialPushdownSymbol)), SingleInitialState(std::move(initialState)) {

}

AutomatonBase* InputDrivenDPDA::clone() const {
	return new InputDrivenDPDA(*this);
}

AutomatonBase* InputDrivenDPDA::plunder() && {
	return new InputDrivenDPDA(std::move(*this));
}

bool InputDrivenDPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& t : transitions) {
		if (t.first.first == state || t.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool InputDrivenDPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.first.second == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool InputDrivenDPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const auto& pushdownStoreOperation : inputSymbolToPushdownStoreOperation) {
		if (std::find(pushdownStoreOperation.second.first.begin(), pushdownStoreOperation.second.first.end(), symbol) != pushdownStoreOperation.second.first.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		if (std::find(pushdownStoreOperation.second.second.begin(), pushdownStoreOperation.second.second.end(), symbol) != pushdownStoreOperation.second.second.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(initialSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

void InputDrivenDPDA::checkPushdownStoreOperation(const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const std::vector<alphabet::Symbol>& push) {
	if (inputAlphabet.find(input) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	for(const alphabet::Symbol& popSymbol : pop) {
		if (stackAlphabet.find(popSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) popSymbol + "\" doesn't exist.");
		}
	}

	for(const alphabet::Symbol& pushSymbol : push) {
		if (stackAlphabet.find(pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) pushSymbol + "\" doesn't exist.");
		}
	}
}

bool InputDrivenDPDA::setPushdownStoreOperation(alphabet::Symbol input, std::vector<alphabet::Symbol> pop, std::vector<alphabet::Symbol> push) {
	checkPushdownStoreOperation(input, pop, push);
	return inputSymbolToPushdownStoreOperation.insert(std::make_pair(std::move(input), std::make_pair(std::move(pop), std::move(push)))).second;
}

void InputDrivenDPDA::setPushdownStoreOperations(std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>> operations) {
	std::set<alphabet::Symbol> removed;
	std::set_difference(inputAlphabet.begin(), inputAlphabet.end(), key_begin(operations), key_end(operations), std::inserter(removed, removed.end()));

	for(const alphabet::Symbol& removedSymbol : removed) {
		clearPushdownStoreOperation(removedSymbol);
	}

	for(const auto& added : operations) {
		checkPushdownStoreOperation(added.first, added.second.first, added.second.second);
	}

	inputSymbolToPushdownStoreOperation = std::move(operations);
}

bool InputDrivenDPDA::clearPushdownStoreOperation(const alphabet::Symbol& input) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.first.second == input)
			throw AutomatonException("Input symbol \"" + (std::string) input + "\" is used.");
	}

	return inputSymbolToPushdownStoreOperation.erase(input);
}

const std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>>& InputDrivenDPDA::getPushdownStoreOperations() const {
	return inputSymbolToPushdownStoreOperation;
}

bool InputDrivenDPDA::addTransition(State from, alphabet::Symbol input, State to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (inputAlphabet.find(input) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (inputSymbolToPushdownStoreOperation.find(input) == inputSymbolToPushdownStoreOperation.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(std::move(from), std::move(input));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == to)
			return false;
		else
			throw AutomatonException(
				"Transition from this state and symbol already exists (\"" + (std::string) key.first.getName() + "\", \"" + (std::string) key.second + "\") -> \"" + (std::string) to.getName()
						+ "\".");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool InputDrivenDPDA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& InputDrivenDPDA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, alphabet::Symbol>, State > InputDrivenDPDA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsFromState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.first.first == from) {
			transitionsFromState.insert(transition);
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, alphabet::Symbol>, State> InputDrivenDPDA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsToState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.second == to) {
			transitionsToState.insert(transition);
		}
	}

	return transitionsToState;
}

int InputDrivenDPDA::compare(const InputDrivenDPDA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, stackAlphabet, initialSymbol, inputSymbolToPushdownStoreOperation, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.initialSymbol, other.inputSymbolToPushdownStoreOperation, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void InputDrivenDPDA::operator>>(std::ostream& out) const {
	out << "(InputDrivenDPDA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialState = " << initialState
		<< " finalStates = " << finalStates
		<< " stackAlphabet = " << stackAlphabet
		<< " initialSymbol = " << initialSymbol
		<< " transitions = " << transitions
		<< " inputSymbolToPushdownStoreOperation = " << inputSymbolToPushdownStoreOperation
		<< ")";
}

InputDrivenDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string InputDrivenDPDA::XML_TAG_NAME = "InputDrivenDPDA";

InputDrivenDPDA InputDrivenDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, InputDrivenDPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol initialStackSymbol = AutomatonFromXMLParser::parseInitialStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	InputDrivenDPDA automaton(std::move(initialState), std::move(initialStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	automaton.setPushdownStoreOperations(AutomatonFromXMLParser::parseInputToPushdownStoreOperation(input));
	AutomatonFromXMLParser::parseTransitions<InputDrivenDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, InputDrivenDPDA::XML_TAG_NAME);
	return automaton;
}

void InputDrivenDPDA::parseTransition(std::deque<sax::Token>::iterator& input, InputDrivenDPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to));
}

void InputDrivenDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(InputDrivenDPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialStackSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	AutomatonToXMLComposer::composeInputToPushdownStoreOperation(out, this->getPushdownStoreOperations());
	composeTransitions(out);

	out.emplace_back(InputDrivenDPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void InputDrivenDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

xmlApi<automaton::Automaton>::ParserRegister<automaton::InputDrivenDPDA> inputDivenDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::InputDrivenDPDA>(automaton::InputDrivenDPDA::XML_TAG_NAME, automaton::InputDrivenDPDA::parse);
xmlApi<alib::Object>::ParserRegister<automaton::InputDrivenDPDA> inputDivenDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::InputDrivenDPDA>(automaton::InputDrivenDPDA::XML_TAG_NAME, automaton::InputDrivenDPDA::parse);

} /* namespace alib */
