/*
 * CFG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "CFG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

CFG::CFG(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {

}

CFG::CFG(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* CFG::clone() const {
	return new CFG(*this);
}

GrammarBase* CFG::plunder() && {
	return new CFG(std::move(*this));
}

bool CFG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool CFG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool CFG::addRule(alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	for(const alphabet::Symbol& symbol : rightHandSide)
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");

	return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
}

const std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> & CFG::getRules() const {
	return rules;
}

bool CFG::addRawRule(alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	return addRule(std::move(leftHandSide), std::move(rightHandSide));
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> CFG::getRawRules() const {
	return rules;
}

bool CFG::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	return removeRule(leftHandSide, rightHandSide);
}

bool CFG::removeRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

int CFG::compare(const CFG& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void CFG::operator>>(std::ostream& out) const {
	out << "(CFG "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

CFG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string CFG::XML_TAG_NAME = "CFG";

CFG CFG::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, CFG::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	CFG grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, CFG::XML_TAG_NAME);
	return grammar;
}

void CFG::parseRule(std::deque<sax::Token>::iterator& input, CFG& grammar) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS(input);
	std::vector<alphabet::Symbol> rhs = GrammarFromXMLParser::parseRuleRHS(input);
	grammar.addRule(std::move(lhs), std::move(rhs));
}

void CFG::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(CFG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);

	out.emplace_back(CFG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void CFG::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleSingleSymbolLHS(out, rule.first);
			GrammarToXMLComposer::composeRuleRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

xmlApi<grammar::Grammar>::ParserRegister<grammar::CFG> CFGParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::CFG>(grammar::CFG::XML_TAG_NAME, grammar::CFG::parse);
xmlApi<alib::Object>::ParserRegister<grammar::CFG> CFGParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::CFG>(grammar::CFG::XML_TAG_NAME, grammar::CFG::parse);

} /* namespace alib */
