/*
 * CSG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "CSG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

CSG::CSG(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)), generatesEpsilon(false) {

}

CSG::CSG(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)), generatesEpsilon(false) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* CSG::clone() const {
	return new CSG(*this);
}

GrammarBase* CSG::plunder() && {
	return new CSG(std::move(*this));
}

bool CSG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		for(const alphabet::Symbol& lCont : std::get<0>(rule.first))
			if(lCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const alphabet::Symbol& rCont : std::get<2>(rule.first))
			if(rCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool CSG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		for(const alphabet::Symbol& lCont : std::get<0>(rule.first))
			if(lCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		if(std::get<1>(rule.first) == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const alphabet::Symbol& rCont : std::get<2>(rule.first))
			if(rCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool CSG::addRule(std::vector<alphabet::Symbol> lContext, alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rContext, std::vector<alphabet::Symbol> rightHandSide) {
	int rSize = rightHandSide.size();

	if(rSize == 0) {
		throw GrammarException("Epsilon rule is not allowed");
	} else {
		for(const alphabet::Symbol& symbol : lContext) {
			if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
		}

		if(!nonterminalAlphabet.count(leftHandSide))
			throw GrammarException("Rule must rewrite nonterminal symbol");

		for(const alphabet::Symbol& symbol : rContext) {
			if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
		}

		for(const alphabet::Symbol& symbol : rightHandSide) {
			if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
		}

		return rules[make_tuple(std::move(lContext), std::move(leftHandSide), std::move(rContext))].insert(std::move(rightHandSide)).second;
	}
}

const std::map<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>> & CSG::getRules() const {
	return rules;
}

bool CSG::removeRule(const std::vector<alphabet::Symbol>& lContext, const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rContext, const std::vector<alphabet::Symbol>& rightHandSide) {
	return rules[make_tuple(lContext, leftHandSide, rContext)].erase(rightHandSide);
}

void CSG::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool CSG::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

int CSG::compare(const CSG& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void CSG::operator>>(std::ostream& out) const {
	out << "(CSG "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< "generatesEpsilon = " << generatesEpsilon
			<< ")";
}

CSG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string CSG::XML_TAG_NAME = "CSG";

CSG CSG::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, CSG::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	CSG grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon(input);
	grammar.setGeneratesEpsilon(generatesEpsilon);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, CSG::XML_TAG_NAME);
	return grammar;
}

void CSG::parseRule(std::deque<sax::Token>::iterator& input, CSG& grammar) {
	std::vector<alphabet::Symbol> lContext = GrammarFromXMLParser::parseRuleLContext(input);
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS(input);
	std::vector<alphabet::Symbol> rContext = GrammarFromXMLParser::parseRuleRContext(input);
	std::vector<alphabet::Symbol> rhs = GrammarFromXMLParser::parseRuleRHS(input);
	grammar.addRule(std::move(lContext), std::move(lhs), std::move(rContext), std::move(rhs));
}

void CSG::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(CSG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);
	GrammarToXMLComposer::composeGeneratesEpsilon(out, this->getGeneratesEpsilon());

	out.emplace_back(CSG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void CSG::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleLContext(out, std::get<0>(rule.first));
			GrammarToXMLComposer::composeRuleSingleSymbolLHS(out, std::get<1>(rule.first));
			GrammarToXMLComposer::composeRuleRContext(out, std::get<2>(rule.first));
			GrammarToXMLComposer::composeRuleRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

xmlApi<grammar::Grammar>::ParserRegister<grammar::CSG> CSGParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::CSG>(grammar::CSG::XML_TAG_NAME, grammar::CSG::parse);
xmlApi<alib::Object>::ParserRegister<grammar::CSG> CSGParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::CSG>(grammar::CSG::XML_TAG_NAME, grammar::CSG::parse);

} /* namespace alib */
