/*
 * NonContractingGrammar.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "NonContractingGrammar.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

NonContractingGrammar::NonContractingGrammar(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

NonContractingGrammar::NonContractingGrammar(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)), generatesEpsilon(false) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* NonContractingGrammar::clone() const {
	return new NonContractingGrammar(*this);
}

GrammarBase* NonContractingGrammar::plunder() && {
	return new NonContractingGrammar(std::move(*this));
}

bool NonContractingGrammar::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool NonContractingGrammar::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool NonContractingGrammar::addRule(std::vector<alphabet::Symbol> leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	int lSize = leftHandSide.size();
	int rSize = rightHandSide.size();

	if(lSize > rSize)
		throw GrammarException("Invalid size of right hand side of a rule");

	if(std::all_of(leftHandSide.begin(), leftHandSide.end(), [&](const alphabet::Symbol symbol) {return !nonterminalAlphabet.count(symbol);}))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	for(const alphabet::Symbol& symbol : leftHandSide)
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");

	for(const alphabet::Symbol& symbol : rightHandSide) {
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
	}

	return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
}

const std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> & NonContractingGrammar::getRules() const {
	return rules;
}

bool NonContractingGrammar::removeRule(const std::vector<alphabet::Symbol>& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

void NonContractingGrammar::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool NonContractingGrammar::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

int NonContractingGrammar::compare(const NonContractingGrammar& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void NonContractingGrammar::operator>>(std::ostream& out) const {
	out << "(NonContractingGrammar "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< "generatesEpsilon = " << generatesEpsilon
			<< ")";
}

NonContractingGrammar::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string NonContractingGrammar::XML_TAG_NAME = "NonContractingGrammar";

NonContractingGrammar NonContractingGrammar::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NonContractingGrammar::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	NonContractingGrammar grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon(input);
	grammar.setGeneratesEpsilon(generatesEpsilon);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NonContractingGrammar::XML_TAG_NAME);
	return grammar;
}

void NonContractingGrammar::parseRule(std::deque<sax::Token>::iterator& input, NonContractingGrammar& grammar) {
	std::vector<alphabet::Symbol> lhs = GrammarFromXMLParser::parseRuleLHS(input);
	std::vector<alphabet::Symbol> rhs = GrammarFromXMLParser::parseRuleRHS(input);
	grammar.addRule(std::move(lhs), std::move(rhs));
}

void NonContractingGrammar::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NonContractingGrammar::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);
	GrammarToXMLComposer::composeGeneratesEpsilon(out, this->getGeneratesEpsilon());

	out.emplace_back(NonContractingGrammar::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void NonContractingGrammar::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleLHS(out, rule.first);
			GrammarToXMLComposer::composeRuleRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

xmlApi<grammar::Grammar>::ParserRegister<grammar::NonContractingGrammar> nonContractingGrammarParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::NonContractingGrammar>(grammar::NonContractingGrammar::XML_TAG_NAME, grammar::NonContractingGrammar::parse);
xmlApi<alib::Object>::ParserRegister<grammar::NonContractingGrammar> nonContractingGrammarParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::NonContractingGrammar>(grammar::NonContractingGrammar::XML_TAG_NAME, grammar::NonContractingGrammar::parse);

} /* namespace alib */
