/*
 * GraphFeatures.h
 *
 *  Created on: Jun 19, 2014
 *      Author: David Rosca
 */

#include "AdjacencyListUndirectedGraph.h"
#include <compare>
#include "utils.h"

namespace graph {

std::set<Node> AdjacencyListUndirectedGraph::allNodes() const
{
	std::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

std::set<UndirectedEdge> AdjacencyListUndirectedGraph::allEdges() const
{
	return edges;
}

bool AdjacencyListUndirectedGraph::hasNode(const Node &node) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyListUndirectedGraph::hasEdge(const UndirectedEdge &edge) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyListUndirectedGraph::hasEdge(const Node &first, const Node &second) const
{
	auto search = adj.find(first);
	if (search == adj.end()) {
		return false;
	}
	return std::find(search->second.begin(), search->second.end(), second) != search->second.end();
}

std::set<UndirectedEdge> AdjacencyListUndirectedGraph::findEdges(const Node &first, const Node &second) const
{
	std::set<UndirectedEdge> out;

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, first, second)) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyListUndirectedGraph::neighbors(const Node &node) const
{
	std::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

std::set<UndirectedEdge> AdjacencyListUndirectedGraph::neighborEdges(const Node &node) const
{
	std::set<UndirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, node, nghbrs)) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyListUndirectedGraph::addNode(const Node &node)
{
	if (adj.find(node) != adj.end()) {
		return false;
	}

	adj[node]; // insert key
	return true;
}

bool AdjacencyListUndirectedGraph::removeNode(const Node &node)
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);

	for (auto& i : edges) {
		if (i.getFirstNode() == node) {
			listRemoveOne(adj[i.getSecondNode()], node);
			edges.erase(i);
		} else if (i.getSecondNode() == node) {
			listRemoveOne(adj[i.getFirstNode()], node);
			edges.erase(i);
		}
	}

	return true;
}

bool AdjacencyListUndirectedGraph::addEdge(const UndirectedEdge &edge)
{
	if (edges.find(edge) != edges.end()) {
		return false;
	}

	edges.insert(edge);

	adj[edge.getFirstNode()].push_back(edge.getSecondNode());
	adj[edge.getSecondNode()].push_back(edge.getFirstNode());

	return true;
}

bool AdjacencyListUndirectedGraph::removeEdge(const UndirectedEdge &edge)
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);

	auto &lst1 = adj[edge.getFirstNode()];
	listRemoveOne(lst1, edge.getSecondNode());
	if (lst1.empty()) {
		adj.erase(edge.getFirstNode());
	}

	auto &lst2 = adj[edge.getSecondNode()];
	listRemoveOne(lst2, edge.getFirstNode());
	if (lst2.empty()) {
		adj.erase(edge.getSecondNode());
	}

	return true;
}

void AdjacencyListUndirectedGraph::copy(IUndirectedGraph *other)
{
	AdjacencyListUndirectedGraph *o = static_cast<AdjacencyListUndirectedGraph*>(other);

	edges = o->edges;
	adj = o->adj;
}

int AdjacencyListUndirectedGraph::compare(IUndirectedGraph *other) const
{
	AdjacencyListUndirectedGraph *o = static_cast<AdjacencyListUndirectedGraph*>(other);

	std::compare<decltype(edges)> comp;
	int res = comp(edges, o->edges);
	if (res != 0) {
		return res;
	}

	if (adj.size() < o->adj.size()) {
		return -1;
	}

	if (adj.size() > o->adj.size()) {
		return 1;
	}

	for (auto i : adj) {
		auto lst1 = i.second;
		auto lst2 = o->adj.at(i.first);

		std::set<Node> adj1;
		std::set<Node> adj2;
		std::copy(lst1.begin(), lst1.end(), std::inserter(adj1, adj1.begin()));
		std::copy(lst2.begin(), lst2.end(), std::inserter(adj2, adj2.begin()));

		std::compare<std::set<Node>> comp;
		int res = comp(adj1, adj2);
		if (res != 0) {
			return res;
		}
	}

	return 0;
}

} // namespace graph
