/*
 * SuffixTrieNodeTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "SuffixTrieNodeTerminatingSymbol.h"
#include "SuffixTrieTerminatingSymbol.h"
#include "../../exception/AlibException.h"
#include <sstream>

namespace indexes {

SuffixTrieNodeTerminatingSymbol::SuffixTrieNodeTerminatingSymbol ( std::map < alphabet::Symbol, SuffixTrieNodeTerminatingSymbol * > children ) : children ( std::move ( children ) ), parentTree ( NULL ) {
	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeTerminatingSymbol::SuffixTrieNodeTerminatingSymbol ( const SuffixTrieNodeTerminatingSymbol & other ) : parentTree ( NULL ) {
	for ( const auto & element : other.children )
		children.insert ( std::make_pair ( element.first, element.second->clone ( ) ) );

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeTerminatingSymbol::SuffixTrieNodeTerminatingSymbol ( SuffixTrieNodeTerminatingSymbol && other ) noexcept : children ( std::move ( other.children ) ), parentTree ( NULL ) {
	other.children.clear ( );

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieNodeTerminatingSymbol::operator =( const SuffixTrieNodeTerminatingSymbol & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieNodeTerminatingSymbol ( other );

	return * this;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieNodeTerminatingSymbol::operator =( SuffixTrieNodeTerminatingSymbol && other ) noexcept {
	std::swap ( this->children, other.children );
	std::swap ( this->parentTree, other.parentTree ); // this->parentTree is stored within other.parentTree and it is reattached on the next line

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( other.parentTree );

	return * this;
}

SuffixTrieNodeTerminatingSymbol::~SuffixTrieNodeTerminatingSymbol ( ) noexcept {
	for ( auto element : children )
		delete element.second;

	children.clear ( );
}

const std::map < const alphabet::Symbol, const SuffixTrieNodeTerminatingSymbol * > & SuffixTrieNodeTerminatingSymbol::getChildren ( ) const {
	return * reinterpret_cast < const std::map < const alphabet::Symbol, const SuffixTrieNodeTerminatingSymbol * > * > ( & children );
}

const std::map < alphabet::Symbol, SuffixTrieNodeTerminatingSymbol * > & SuffixTrieNodeTerminatingSymbol::getChildren ( ) {
	return children;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieNodeTerminatingSymbol::getChild ( const alphabet::Symbol & symbol ) {
	std::map < alphabet::Symbol, SuffixTrieNodeTerminatingSymbol * >::const_iterator iter = children.find ( symbol );

	if ( iter == children.end ( ) ) throw exception::AlibException ( "child does not exist" );

	return * iter->second;
}

const SuffixTrieNodeTerminatingSymbol & SuffixTrieNodeTerminatingSymbol::getChild ( const alphabet::Symbol & symbol ) const {
	std::map < alphabet::Symbol, SuffixTrieNodeTerminatingSymbol * >::const_iterator iter = children.find ( symbol );

	if ( iter == children.end ( ) ) throw exception::AlibException ( "child does not exist" );

	return * iter->second;
}

bool SuffixTrieNodeTerminatingSymbol::hasChild ( const alphabet::Symbol & symbol ) const {
	if ( children.find ( symbol ) == children.end ( ) ) return false;

	return true;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieNodeTerminatingSymbol::addChild ( alphabet::Symbol symbol, SuffixTrieNodeTerminatingSymbol node ) {
	std::map < alphabet::Symbol, SuffixTrieNodeTerminatingSymbol * >::iterator iter = children.find ( symbol );

	if ( iter != children.end ( ) ) throw exception::AlibException ( "child already exist" );

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) throw exception::AlibException ( "Symbol is not in the alphabet" );

	return * children.insert ( std::make_pair ( std::move ( symbol ), new SuffixTrieNodeTerminatingSymbol ( std::move ( node ) ) ) ).first->second;
}

SuffixTrieNodeTerminatingSymbol * SuffixTrieNodeTerminatingSymbol::getParent ( ) {
	return parent;
}

const SuffixTrieNodeTerminatingSymbol * SuffixTrieNodeTerminatingSymbol::getParent ( ) const {
	return parent;
}

void SuffixTrieNodeTerminatingSymbol::swap ( SuffixTrieNodeTerminatingSymbol & other ) {
	const TerminatingSymbolAlphabet * thisParentTree  = this->parentTree;
	const TerminatingSymbolAlphabet * otherParentTree = other.parentTree;

	SuffixTrieNodeTerminatingSymbol tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachTree ( thisParentTree );
	other.attachTree ( otherParentTree );
}

SuffixTrieNodeTerminatingSymbol * SuffixTrieNodeTerminatingSymbol::clone ( ) const {
	return new SuffixTrieNodeTerminatingSymbol ( * this );
}

SuffixTrieNodeTerminatingSymbol * SuffixTrieNodeTerminatingSymbol::plunder ( ) && {
	return new SuffixTrieNodeTerminatingSymbol ( std::move ( * this ) );
}

int SuffixTrieNodeTerminatingSymbol::compare ( const SuffixTrieNodeTerminatingSymbol & other ) const {
	int thisSize  = this->children.size ( );
	int otherSize = other.children.size ( );

	if ( thisSize < otherSize ) return -1;

	if ( thisSize > otherSize ) return 1;

	auto thisIter  = this->children.begin ( );
	auto otherIter = other.children.begin ( );

	for ( ; thisIter != this->children.end ( ); ++thisIter, ++otherIter ) {
		int res = thisIter->first.compare ( otherIter->first );

		if ( res != 0 ) return res;

		res = thisIter->second->compare ( * otherIter->second );

		if ( res != 0 ) return res;
	}

	return 0;
}

void SuffixTrieNodeTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieNodeTerminatingSymbol " << " children = " << this->children << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieNodeTerminatingSymbol & node ) {
	node >> out;
	return out;
}

bool SuffixTrieNodeTerminatingSymbol::testSymbol ( const alphabet::Symbol & symbol ) const {
	for ( const auto & child : this->children ) {
		if ( symbol == child.first ) return true;

		if ( child.second->testSymbol ( symbol ) ) return true;
	}

	return false;
}

bool SuffixTrieNodeTerminatingSymbol::attachTree ( const TerminatingSymbolAlphabet * tree ) {
	if ( this->parentTree == tree ) return true;

	this->parentTree = tree;

	for ( const auto & child : this->children ) {
		if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( child.first ) == this->parentTree->getAlphabet ( ).end ( ) ) ) return false;

		if ( !child.second->attachTree ( tree ) ) return false;
	}

	return true;
}

void SuffixTrieNodeTerminatingSymbol::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	for ( const auto & child : this->children ) {
		alphabet.insert ( child.first );
		child.second->computeMinimalAlphabet ( alphabet );
	}
}

SuffixTrieNodeTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace indexes */
