/*
 * RegExpFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpFromXMLParser.h"
#include "../../sax/ParserException.h"

#include "../RegExp.h"
#include "../RegExpClasses.h"

#include "../unbounded/UnboundedRegExpElements.h"
#include "../formal/FormalRegExpElements.h"
#include "../../XmlApi.hpp"

namespace regexp {

std::set<alphabet::Symbol> RegExpFromXMLParser::parseAlphabet(std::deque<sax::Token>::iterator& input) {
	std::set<alphabet::Symbol> alphabet;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "alphabet");
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		alphabet.insert(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "alphabet");
	return alphabet;
}

UnboundedRegExpElement* RegExpFromXMLParser::parseUnboundedRegExpElement(std::deque<sax::Token>::iterator& input) {
	if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "empty")) {
		return parseUnboundedRegExpEmpty(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		return parseUnboundedRegExpEpsilon(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "iteration")) {
		return parseUnboundedRegExpIteration(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "alternation")) {
		return parseUnboundedRegExpAlternation(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation")) {
		return parseUnboundedRegExpConcatenation(input);
	} else {
		return new UnboundedRegExpSymbol(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
}

UnboundedRegExpAlternation* RegExpFromXMLParser::parseUnboundedRegExpAlternation(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "alternation");

	UnboundedRegExpAlternation* alternation = new UnboundedRegExpAlternation();

	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
		alternation->appendElement(std::move(*element));
		delete element;
	}

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "alternation");
	return alternation;
}

UnboundedRegExpConcatenation* RegExpFromXMLParser::parseUnboundedRegExpConcatenation(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation");

	UnboundedRegExpConcatenation* concatenation = new UnboundedRegExpConcatenation();

	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
		concatenation->appendElement(std::move(*element));
		delete element;
	}

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "concatenation");
	return concatenation;

}

UnboundedRegExpIteration* RegExpFromXMLParser::parseUnboundedRegExpIteration(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "iteration");

	UnboundedRegExpElement* element = parseUnboundedRegExpElement(input);
	UnboundedRegExpIteration* iteration = new UnboundedRegExpIteration(std::move(*element));
	delete element;

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "iteration");
	return iteration;
}

UnboundedRegExpEpsilon* RegExpFromXMLParser::parseUnboundedRegExpEpsilon(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon");

	UnboundedRegExpEpsilon* epsilon = new UnboundedRegExpEpsilon();

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	return epsilon;
}

UnboundedRegExpEmpty* RegExpFromXMLParser::parseUnboundedRegExpEmpty(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "empty");

	UnboundedRegExpEmpty* empty = new UnboundedRegExpEmpty();

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "empty");
	return empty;
}

FormalRegExpElement* RegExpFromXMLParser::parseFormalRegExpElement(std::deque<sax::Token>::iterator& input) {
	if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "empty")) {
		return parseFormalRegExpEmpty(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		return parseFormalRegExpEpsilon(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "iteration")) {
		return parseFormalRegExpIteration(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "alternation")) {
		return parseFormalRegExpAlternation(input);
	} else if (sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation")) {
		return parseFormalRegExpConcatenation(input);
	} else {
		return new FormalRegExpSymbol(alib::xmlApi<alphabet::Symbol>::parse(input));
	}
}

FormalRegExpAlternation* RegExpFromXMLParser::parseFormalRegExpAlternation(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "alternation");

	FormalRegExpElement* element1 = parseFormalRegExpElement(input);
	FormalRegExpElement* element2 = parseFormalRegExpElement(input);

	FormalRegExpAlternation* alternation = new FormalRegExpAlternation(std::move(*element1), std::move(*element2));

	delete element1;
	delete element2;

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "alternation");
	return alternation;
}

FormalRegExpConcatenation* RegExpFromXMLParser::parseFormalRegExpConcatenation(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation");

	FormalRegExpElement* element1 = parseFormalRegExpElement(input);
	FormalRegExpElement* element2 = parseFormalRegExpElement(input);

	FormalRegExpConcatenation* concatenation = new FormalRegExpConcatenation(std::move(*element1), std::move(*element2));

	delete element1;
	delete element2;

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "concatenation");
	return concatenation;

}

FormalRegExpIteration* RegExpFromXMLParser::parseFormalRegExpIteration(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "iteration");

	FormalRegExpElement* element = parseFormalRegExpElement(input);
	FormalRegExpIteration* iteration = new FormalRegExpIteration(std::move(*element));
	delete element;

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "iteration");
	return iteration;
}

FormalRegExpEpsilon* RegExpFromXMLParser::parseFormalRegExpEpsilon(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon");

	FormalRegExpEpsilon* epsilon = new FormalRegExpEpsilon();

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	return epsilon;
}

FormalRegExpEmpty* RegExpFromXMLParser::parseFormalRegExpEmpty(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "empty");

	FormalRegExpEmpty* empty = new FormalRegExpEmpty();

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "empty");
	return empty;
}

} /* namespace regexp */

