/*
 * UnboundedRegExpAlternation.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "UnboundedRegExpAlternation.h"
#include "../../exception/AlibException.h"
#include "../formal/FormalRegExpAlternation.h"
#include "../formal/FormalRegExpEmpty.h"
#include <sstream>

namespace regexp {

UnboundedRegExpAlternation::UnboundedRegExpAlternation() {

}

UnboundedRegExpAlternation::UnboundedRegExpAlternation(const UnboundedRegExpAlternation& other) {
	for (const auto& element : other.elements) {
		elements.push_back(element->clone());
	}
}

UnboundedRegExpAlternation::UnboundedRegExpAlternation(UnboundedRegExpAlternation&& other) noexcept : elements(std::move(other.elements)) {
	other.elements.clear();
	this->attachRegExp(NULL);
}

UnboundedRegExpAlternation& UnboundedRegExpAlternation::operator=(const UnboundedRegExpAlternation& other) {
	if (this == &other) {
		return *this;
	}

	*this = UnboundedRegExpAlternation(other);

	return *this;
}

UnboundedRegExpAlternation& UnboundedRegExpAlternation::operator=(UnboundedRegExpAlternation&& other) noexcept {
	std::swap(this->elements, other.elements);
	std::swap(this->parentRegExp, other.parentRegExp); // this->parentRegExp is stored within other.parentRegExp and it is reattached on the next line, done to save code (loop is in attachRegExp method)

	this->attachRegExp(other.parentRegExp);

	return *this;
}

UnboundedRegExpAlternation::~UnboundedRegExpAlternation() noexcept {
	for (auto element : elements) {
		delete element;
	}
	elements.clear();
}

const std::vector<const UnboundedRegExpElement*> & UnboundedRegExpAlternation::getElements() const {
	return * reinterpret_cast<const std::vector<const UnboundedRegExpElement*> * > (&elements);
}

const std::vector<UnboundedRegExpElement*> & UnboundedRegExpAlternation::getElements() {
	return elements;
}

void UnboundedRegExpAlternation::appendElement(const UnboundedRegExpElement& element) {
	UnboundedRegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	this->elements.push_back(elem);
}

void UnboundedRegExpAlternation::appendElement(UnboundedRegExpElement&& element) {
	UnboundedRegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	this->elements.push_back(elem);
}

UnboundedRegExpElement* UnboundedRegExpAlternation::clone() const {
	return new UnboundedRegExpAlternation(*this);
}

UnboundedRegExpElement* UnboundedRegExpAlternation::plunder() && {
	return new UnboundedRegExpAlternation(std::move(*this));
}

FormalRegExpElement* UnboundedRegExpAlternation::cloneAsFormal() const {
	if(elements.size() == 0) return new FormalRegExpEmpty();
	if(elements.size() == 1) return elements[0]->cloneAsFormal();

	int size = elements.size();
	FormalRegExpElement* rightTmp = elements[--size]->cloneAsFormal();
	FormalRegExpElement* leftTmp = elements[--size]->cloneAsFormal();

	FormalRegExpAlternation* res = new FormalRegExpAlternation(std::move(*leftTmp), std::move(*rightTmp));
	delete rightTmp;
	delete leftTmp;

	while(size) {
		FormalRegExpElement* tmp = res;
		FormalRegExpElement* newLeft = elements[--size]->cloneAsFormal();

		res = new FormalRegExpAlternation(std::move(*newLeft), std::move(*tmp));
		delete tmp;
		delete newLeft;
	}

	return res;
}

int UnboundedRegExpAlternation::compare(const UnboundedRegExpAlternation& other) const {
	int thisSize = this->elements.size();
	int otherSize = other.elements.size();
	if(thisSize < otherSize) return -1;
	if(thisSize > otherSize) return 1;

	auto thisIter = this->elements.begin();
	auto otherIter = other.elements.begin();
	for(; thisIter != this->elements.end(); ++thisIter, ++otherIter) {
		int res = (*thisIter)->compare(**otherIter);
		if(res != 0) return res;
	}
	return 0;
}

void UnboundedRegExpAlternation::operator>>(std::ostream& out) const {
	out << "(UnboundedRegExpAlternation";
	for(const auto& e : elements)
		out << " " << *e;

	out << ")";
}

bool UnboundedRegExpAlternation::testSymbol( const alphabet::Symbol & symbol ) const {
	for(const auto& child : this->elements)
		if(child->testSymbol(symbol)) return true;
	return false;
}

bool UnboundedRegExpAlternation::attachRegExp(const RegExpAlphabet * regexp ) {
	if(this->parentRegExp == regexp) return true;

	this->parentRegExp = regexp;
	for(const auto& child : this->elements)
		if(!child->attachRegExp(regexp)) return false;
	return true;
}

void UnboundedRegExpAlternation::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	for(const auto& child : this->elements)
		child->computeMinimalAlphabet(alphabet);
}

UnboundedRegExpAlternation::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace regexp */
