/*
 * LinearString.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LinearString.h"
#include "Epsilon.h"
#include "../tree/ranked/PrefixRankedTree.h"
#include "../exception/AlibException.h"

#include <sstream>
#include <algorithm>

#include "../sax/FromXMLParserHelper.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "String.h"
#include "../object/Object.h"
#include "../XmlApi.hpp"
#include "../CastApi.hpp"

namespace string {

LinearString::LinearString ( ) {
}

LinearString::LinearString ( std::set < alphabet::Symbol > alphabet, std::vector < alphabet::Symbol > data ) {
	this->alphabet = std::move ( alphabet );
	setContent ( std::move ( data ) );
}

LinearString::LinearString ( std::vector < alphabet::Symbol > data ) {
	alphabet = std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) );
	m_Data = std::move ( data );
}

LinearString::LinearString ( const std::string & string ) {
	for ( const char & symbol : string )
		m_Data.push_back ( alphabet::symbolFrom ( symbol ) );

	alphabet = std::set < alphabet::Symbol > ( m_Data.begin ( ), m_Data.end ( ) );
}

LinearString::LinearString ( const char * string ) : LinearString ( ( std::string ) string ) {
}

LinearString::LinearString ( const Epsilon & epsilon ) {
	alphabet = epsilon.getAlphabet ( );
}

LinearString::LinearString ( const tree::PrefixRankedTree & tree ) {
	for ( const alphabet::RankedSymbol & symbol : tree.getAlphabet ( ) )
		this->alphabet.insert ( alphabet::Symbol ( std::move ( symbol ) ) );

	for ( const alphabet::RankedSymbol & symbol : tree.getContent ( ) )
		this->m_Data.push_back ( alphabet::Symbol ( symbol ) );
}

StringBase * LinearString::clone ( ) const {
	return new LinearString ( * this );
}

StringBase * LinearString::plunder ( ) && {
	return new LinearString ( std::move ( * this ) );
}

bool LinearString::removeSymbolFromAlphabet ( const alphabet::Symbol & symbol ) {
	if ( std::any_of ( m_Data.begin ( ), m_Data.end ( ), [&] ( const alphabet::Symbol & s ) {
			return s == symbol;
		} ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

void LinearString::appendSymbol ( alphabet::Symbol symbol ) {
	if ( alphabet.count ( symbol ) == 0 )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

const std::vector < alphabet::Symbol > & LinearString::getContent ( ) const {
	return this->m_Data;
}

void LinearString::setContent ( std::vector < alphabet::Symbol > data ) {
	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), alphabet.begin ( ), alphabet.end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::AlibException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

bool LinearString::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int LinearString::compare ( const LinearString & other ) const {
	auto first	= std::tie ( m_Data, alphabet );
	auto second = std::tie ( other.m_Data, other.alphabet );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void LinearString::operator >>( std::ostream & out ) const {
	out << "(LinearString ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

LinearString::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

const std::string LinearString::XML_TAG_NAME = "LinearString";

LinearString LinearString::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearString::XML_TAG_NAME );
	std::set < alphabet::Symbol > alphabet = StringFromXMLParser::parseAlphabet ( input );
	std::vector < alphabet::Symbol > content = StringFromXMLParser::parseContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearString::XML_TAG_NAME );

	return LinearString ( alphabet, content );
}

void LinearString::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearString::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, alphabet );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearString::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace string */

namespace alib {

xmlApi < string::String >::ParserRegister < string::LinearString > linearStringParserRegister = xmlApi < string::String >::ParserRegister < string::LinearString > ( string::LinearString::XML_TAG_NAME, string::LinearString::parse );
xmlApi < alib::Object >::ParserRegister < string::LinearString > linearStringParserRegister2  = xmlApi < alib::Object >::ParserRegister < string::LinearString > ( string::LinearString::XML_TAG_NAME, string::LinearString::parse );

auto LinearStringFromEpsilon = castApi::CastRegister < string::LinearString, string::Epsilon > ( );
auto LinearStringFromPrefixRankedTree = castApi::CastRegister < string::LinearString, tree::PrefixRankedTree > ( );
auto LinearStringCastBinder = castApi::CastPoolStringBinder < string::LinearString > ( string::LinearString::XML_TAG_NAME );

} /* namespace alib */
