/*
 * LinearStringTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LinearStringTerminatingSymbol.h"
#include "LinearString.h"
#include "../exception/AlibException.h"

#include <sstream>
#include <algorithm>

#include "../sax/FromXMLParserHelper.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "String.h"
#include "../object/Object.h"
#include "../XmlApi.hpp"
#include "../CastApi.hpp"

namespace string {

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol ) : TerminatingSymbolStringAlphabet ( std::move ( terminatingSymbol ) ) {
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data ) : TerminatingSymbolStringAlphabet ( std::move ( terminatingSymbol ) ) {
	this->alphabet = std::move ( alphabet );
	setContent ( std::move ( data ) );
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data ) : TerminatingSymbolStringAlphabet ( std::move ( terminatingSymbol ) ) {
	alphabet = std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) );
	m_Data = std::move ( data );
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, const LinearString & string ) : TerminatingSymbolStringAlphabet ( std::move ( terminatingSymbol ) ) {
	alphabet = string.getAlphabet ( );
	m_Data = string.getContent ( );
}

StringBase * LinearStringTerminatingSymbol::clone ( ) const {
	return new LinearStringTerminatingSymbol ( * this );
}

StringBase * LinearStringTerminatingSymbol::plunder ( ) && {
	return new LinearStringTerminatingSymbol ( std::move ( * this ) );
}

bool LinearStringTerminatingSymbol::removeSymbolFromAlphabet ( const alphabet::Symbol & symbol ) {
	if ( std::any_of ( m_Data.begin ( ), m_Data.end ( ), [&] ( const alphabet::Symbol & s ) {
			return s == symbol;
		} ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

void LinearStringTerminatingSymbol::appendSymbol ( alphabet::Symbol symbol ) {
	if ( alphabet.count ( symbol ) == 0 )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

const std::vector < alphabet::Symbol > & LinearStringTerminatingSymbol::getContent ( ) const {
	return this->m_Data;
}

void LinearStringTerminatingSymbol::setContent ( std::vector < alphabet::Symbol > data ) {
	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), alphabet.begin ( ), alphabet.end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::AlibException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

bool LinearStringTerminatingSymbol::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int LinearStringTerminatingSymbol::compare ( const LinearStringTerminatingSymbol & other ) const {
	auto first	= std::tie ( m_Data, alphabet );
	auto second = std::tie ( other.m_Data, other.alphabet );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void LinearStringTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(LinearStringTerminatingSymbol ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

LinearStringTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

const std::string LinearStringTerminatingSymbol::XML_TAG_NAME = "LinearStringTerminatingSymbol";

LinearStringTerminatingSymbol LinearStringTerminatingSymbol::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearStringTerminatingSymbol::XML_TAG_NAME );
	std::set < alphabet::Symbol > alphabet = StringFromXMLParser::parseAlphabet ( input );
	alphabet::Symbol terminatingSymbol = alib::xmlApi < alphabet::Symbol >::parse ( input );
	std::vector < alphabet::Symbol > content = StringFromXMLParser::parseContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearStringTerminatingSymbol::XML_TAG_NAME );

	return LinearStringTerminatingSymbol ( alphabet, terminatingSymbol, content );
}

void LinearStringTerminatingSymbol::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearStringTerminatingSymbol::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, alphabet );
	alib::xmlApi < alphabet::Symbol >::compose ( out, terminatingSymbol );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearStringTerminatingSymbol::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace string */

namespace alib {

xmlApi < string::String >::ParserRegister < string::LinearStringTerminatingSymbol > linearStringTerminatingSymbolParserRegister = xmlApi < string::String >::ParserRegister < string::LinearStringTerminatingSymbol > ( string::LinearStringTerminatingSymbol::XML_TAG_NAME, string::LinearStringTerminatingSymbol::parse );
xmlApi < alib::Object >::ParserRegister < string::LinearStringTerminatingSymbol > linearStringTerminatingSymbolParserRegister2	= xmlApi < alib::Object >::ParserRegister < string::LinearStringTerminatingSymbol > ( string::LinearStringTerminatingSymbol::XML_TAG_NAME, string::LinearStringTerminatingSymbol::parse );

} /* namespace alib */
