/*
 * PrefixRankedBarPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixRankedBarPattern.h"
#include "../../exception/AlibException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "RankedPattern.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace tree {

PrefixRankedBarPattern::PrefixRankedBarPattern(alphabet::Symbol bar, alphabet::Symbol variablesBar, alphabet::RankedSymbol subtreeWildcard, std::set<alphabet::RankedSymbol> alphabet, std::vector<alphabet::RankedSymbol> data) : RankedPatternAlphabet(std::move(subtreeWildcard)), bar(std::move(bar)), variablesBar(std::move(variablesBar)) {
	this->alphabet = std::move(alphabet);
	setContent(std::move(data));
}

PrefixRankedBarPattern::PrefixRankedBarPattern(alphabet::Symbol bar, alphabet::Symbol variablesBar, alphabet::RankedSymbol subtreeWildcard, std::vector<alphabet::RankedSymbol> data) : RankedPatternAlphabet(std::move(subtreeWildcard)), bar(std::move(bar)), variablesBar(std::move(variablesBar)) {
	arityChecksum(data);

	alphabet = std::set<alphabet::RankedSymbol>(data.begin(), data.end());
	m_Data = std::move(data);
}

PrefixRankedBarPattern::PrefixRankedBarPattern(alphabet::Symbol bar, alphabet::Symbol variablesBar, const RankedPattern& tree) : RankedPatternAlphabet(tree.getSubtreeWildcard()), bar(std::move(bar)), variablesBar(std::move(variablesBar)) {
	toPrefixRankedBar(tree.getRoot());
	alphabet = std::set<alphabet::RankedSymbol>(m_Data.begin(), m_Data.end());
}

void PrefixRankedBarPattern::toPrefixRankedBar(const RankedNode& node) {
	if(node.getSymbol() == subtreeWildcard) {
		m_Data.push_back(node.getSymbol());
		m_Data.push_back(alphabet::RankedSymbol(variablesBar, node.getSymbol().getRank()));
	} else {
		m_Data.push_back(node.getSymbol());
		for(const RankedNode* child : node.getChildren())
			toPrefixRankedBar(*child);
		m_Data.push_back(alphabet::RankedSymbol(bar, node.getSymbol().getRank()));
	}
}

TreeBase* PrefixRankedBarPattern::clone() const {
	return new PrefixRankedBarPattern(*this);
}

TreeBase* PrefixRankedBarPattern::plunder() && {
	return new PrefixRankedBarPattern(std::move(*this));
}

bool PrefixRankedBarPattern::removeSymbolFromAlphabet(const alphabet::RankedSymbol & symbol) {
	if(std::any_of(m_Data.begin(), m_Data.end(), [&](const alphabet::RankedSymbol & s) { return s == symbol; } ) )
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	if(this->subtreeWildcard == symbol)
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is subtreeWildcard.");

	return alphabet.erase(symbol);
}

const std::vector<alphabet::RankedSymbol>& PrefixRankedBarPattern::getContent() const {
	return this->m_Data;
}

void PrefixRankedBarPattern::setContent(std::vector<alphabet::RankedSymbol> data) {
	arityChecksum(data);

	std::set<alphabet::RankedSymbol> minimalAlphabet(data.begin(), data.end());
	std::set<alphabet::RankedSymbol> unknownSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), alphabet.begin(), alphabet.end(), std::inserter(unknownSymbols, unknownSymbols.end()));

	if(unknownSymbols.size() > 0)
		throw exception::AlibException("Input symbols not in the alphabet.");

	this->m_Data = std::move(data);
}

const alphabet::Symbol& PrefixRankedBarPattern::getBarSymbol() const {
	return bar;
}

const alphabet::Symbol& PrefixRankedBarPattern::getVariablesBarSymbol() const {
	return variablesBar;
}

void PrefixRankedBarPattern::arityChecksum(const std::vector<alphabet::RankedSymbol>& data) {
	int arityChecksumTerminals = 1;
	int arityChecksumBars = 1;
	int arityChecksumTypes = 0;
	for(const alphabet::RankedSymbol& symbol : data) {
		if(symbol.getSymbol() != bar && symbol.getSymbol() != variablesBar) {
			arityChecksumTerminals += symbol.getRank().getData();
			arityChecksumTerminals -= 1;
			arityChecksumTypes += 1;
		} else {
			arityChecksumBars += symbol.getRank().getData();
			arityChecksumBars -= 1;
			arityChecksumTypes -= 1;
		}
	}

	if(arityChecksumTerminals != 0 || arityChecksumBars != 0 || arityChecksumTypes != 0) throw exception::AlibException("The string does not form a tree");

	for(unsigned i = 0; i < data.size(); ++i) {
		if(data[i] == subtreeWildcard && ((i + 1) == data.size() || data[i + 1].getSymbol() != variablesBar)) throw exception::AlibException("Inconsystency of SubtreeWildcard and variablesBar");
	}
}

bool PrefixRankedBarPattern::isEmpty() const {
	return this->m_Data.size() == 0;
}

int PrefixRankedBarPattern::compare(const PrefixRankedBarPattern& other) const {
	auto first = std::tie(m_Data, alphabet);
	auto second = std::tie(other.m_Data, other.alphabet);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void PrefixRankedBarPattern::operator >>(std::ostream& out) const {
	if( this->isEmpty() ) {
		out << "(Epsilon)";
	} else {
		out << "(PrefixRankedBarPattern ";
		for(const alphabet::RankedSymbol& symbol : this->m_Data)
			out << symbol;
		out << ")";
	}
}

PrefixRankedBarPattern::operator std::string () const {
	std::stringstream ss;
	if( this->isEmpty() ) {
		ss << "E";
	} else {
		ss << "\"";
		for(const alphabet::RankedSymbol& symbol : this->m_Data)
			ss << (std::string) symbol;
		ss << "\"";
	}
	return std::move(ss).str();
}

const std::string PrefixRankedBarPattern::XML_TAG_NAME = "PrefixRankedBarPattern";

PrefixRankedBarPattern PrefixRankedBarPattern::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, PrefixRankedBarPattern::XML_TAG_NAME);
	alphabet::Symbol bar = TreeFromXMLParser::parseBar(input);
	alphabet::Symbol variablesBarSymbol = TreeFromXMLParser::parseVariablesBar(input);
	alphabet::RankedSymbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardRankedSymbol(input);
	std::set<alphabet::RankedSymbol> rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet(input);
	std::vector<alphabet::RankedSymbol> data = TreeFromXMLParser::parseContentData(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, PrefixRankedBarPattern::XML_TAG_NAME);
	return PrefixRankedBarPattern(std::move(bar), std::move(variablesBarSymbol), std::move(subtreeWildcardSymbol), std::move(rankedAlphabet), std::move(data));
}

void PrefixRankedBarPattern::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(PrefixRankedBarPattern::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeBar(out, bar);
	TreeToXMLComposer::composeVariablesBar(out, variablesBar);
	TreeToXMLComposer::composeSubtreeWildcard(out, subtreeWildcard);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeContent(out, m_Data);
	out.emplace_back(PrefixRankedBarPattern::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::PrefixRankedBarPattern> prefixRankedBarPatternParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::PrefixRankedBarPattern>(tree::PrefixRankedBarPattern::XML_TAG_NAME, tree::PrefixRankedBarPattern::parse);
xmlApi<alib::Object>::ParserRegister<tree::PrefixRankedBarPattern> prefixRankedBarPatternParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::PrefixRankedBarPattern>(tree::PrefixRankedBarPattern::XML_TAG_NAME, tree::PrefixRankedBarPattern::parse);

} /* namespace alib */
