/*
 * RankedNode.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "RankedNode.h"
#include "RankedTree.h"
#include "../TreeException.h"
#include "../unranked/UnrankedNode.h"
#include <sstream>

namespace tree {

RankedNode::RankedNode ( alphabet::RankedSymbol symbol, std::vector < RankedNode * > children ) : symbol ( std::move ( symbol ) ), children ( std::move ( children ) ), parentTree ( NULL ) {
	if ( this->children.size ( ) != this->symbol.getRank ( ).getData ( ) ) throw TreeException ( "Number of children doesn't match the rank of the symbol" );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

RankedNode::RankedNode ( const RankedNode & other ) : symbol ( other.symbol ), parentTree ( NULL ) {
	for ( const auto & element : other.children )
		children.push_back ( element->clone ( ) );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

RankedNode::RankedNode ( RankedNode && other ) noexcept : symbol ( std::move ( other.symbol ) ), children ( std::move ( other.children ) ), parentTree ( NULL ) {
	other.children.clear ( );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

RankedNode & RankedNode::operator =( const RankedNode & other ) {
	if ( this == & other )
		return * this;

	* this = RankedNode ( other );

	return * this;
}

RankedNode & RankedNode::operator =( RankedNode && other ) noexcept {
	std::swap ( this->symbol, other.symbol );
	std::swap ( this->children, other.children );
	std::swap ( this->parentTree, other.parentTree ); // this->parentTree is stored within other.parentTree and it is reattached on the next line

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( other.parentTree );

	return * this;
}

RankedNode::~RankedNode ( ) noexcept {
	for ( auto element : children )
		delete element;

	children.clear ( );
}

const std::vector < const RankedNode * > & RankedNode::getChildren ( ) const {
	return * reinterpret_cast < const std::vector < const RankedNode * > * > ( & children );
}

const std::vector < RankedNode * > & RankedNode::getChildren ( ) {
	return children;
}

RankedNode * RankedNode::getParent ( ) {
	return parent;
}

const RankedNode * RankedNode::getParent ( ) const {
	return parent;
}

void RankedNode::setSymbol ( alphabet::RankedSymbol symbol ) {
	if ( symbol.getRank ( ) != this->symbol.getRank ( ) ) throw TreeException ( "Number of children doesn't match the rank of the symbol" );

	this->symbol = std::move ( symbol );

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( this->symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) throw TreeException ( "Symbol is not in the alphabet" );
}

void RankedNode::swap ( RankedNode & other ) {
	const RankedAlphabet * thisParentTree  = this->parentTree;
	const RankedAlphabet * otherParentTree = other.parentTree;

	RankedNode tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachTree ( thisParentTree );
	other.attachTree ( otherParentTree );
}

RankedNode * RankedNode::clone ( ) const {
	return new RankedNode ( * this );
}

RankedNode * RankedNode::plunder ( ) && {
	return new RankedNode ( std::move ( * this ) );
}

UnrankedNode * RankedNode::cloneAsUnranked ( ) const {
	std::vector < UnrankedNode * > clonedChildren;

	for ( const RankedNode * child : this->children )
		clonedChildren.push_back ( child->cloneAsUnranked ( ) );

	return new UnrankedNode ( this->symbol.getSymbol ( ), std::move ( clonedChildren ) );
}

int RankedNode::compare ( const RankedNode & other ) const {
	int res = this->symbol.compare ( other.symbol );

	if ( res != 0 ) return res;

	int thisSize  = this->children.size ( );
	int otherSize = other.children.size ( );

	if ( thisSize < otherSize ) return -1;

	if ( thisSize > otherSize ) return 1;

	auto thisIter  = this->children.begin ( );
	auto otherIter = other.children.begin ( );

	for ( ; thisIter != this->children.end ( ); ++thisIter, ++otherIter ) {
		int res = ( * thisIter )->compare ( * * otherIter );

		if ( res != 0 ) return res;
	}

	return 0;
}

void RankedNode::operator >>( std::ostream & out ) const {
	out << "(RankedNode " << " symbol = " << this->symbol << " children = " << this->children << "})";
}

std::ostream & operator <<( std::ostream & out, const RankedNode & node ) {
	node >> out;
	return out;
}

bool RankedNode::testSymbol ( const alphabet::RankedSymbol & symbol ) const {
	if ( symbol == this->symbol ) return true;

	for ( const auto & child : this->children )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool RankedNode::attachTree ( const RankedAlphabet * tree ) {
	if ( this->parentTree == tree ) return true;

	this->parentTree = tree;

	for ( const auto & child : this->children )
		if ( !child->attachTree ( tree ) ) return false;

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( this->symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) return false;

	return true;
}

void RankedNode::computeMinimalAlphabet ( std::set < alphabet::RankedSymbol > & alphabet ) const {
	alphabet.insert ( this->symbol );

	for ( const auto & child : this->children )
		child->computeMinimalAlphabet ( alphabet );
}

const alphabet::RankedSymbol & RankedNode::getSymbol ( ) const {
	return this->symbol;
}

RankedNode::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

void RankedNode::nicePrint ( std::ostream & os, const std::string & prefix, const bool last ) const {
	os << prefix;

	std::string nextPrefix ( prefix );

	if ( last ) {
		os << "\\-";
		nextPrefix += "  ";
	} else {
		os << "|-";
		nextPrefix += "| ";
	}

	os << ( std::string ) symbol.getSymbol ( ) << "(" << symbol.getRank ( ).getData ( ) << ")" << std::endl;

	for ( unsigned int i = 0; i < children.size ( ); ++i ) {
		os << nextPrefix << "|" << std::endl;
		children[i]->nicePrint ( os, nextPrefix, i == children.size ( ) - 1 );
	}
}

} /* namespace tree */
