/*
 * RankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "RankedTree.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedTree.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace tree {

RankedTree::RankedTree(const UnrankedTree& other) {
	this->tree = NULL;
	RankedNode* element = other.getRoot().cloneAsRanked();
	element->computeMinimalAlphabet(alphabet);
	setTree(std::move(*element));
	delete element;
}

RankedTree::RankedTree(std::set<alphabet::RankedSymbol> alphabet, RankedNode tree) {
	this->alphabet = std::move(alphabet);
	this->tree = NULL;
	setTree(std::move(tree));
}

RankedTree::RankedTree(RankedNode tree) {
	tree.computeMinimalAlphabet(alphabet);
	this->tree = NULL;
	setTree(std::move(tree));
}

RankedTree::RankedTree(const RankedTree& other) : tree(other.tree->clone()) {
	alphabet = other.alphabet;
	this->tree->attachTree(this);
}

RankedTree::RankedTree(RankedTree&& other) noexcept : tree(other.tree) {
	alphabet = std::move(other.alphabet);
	this->tree->attachTree(this);
	other.tree = NULL;
}

TreeBase* RankedTree::clone() const {
	return new RankedTree(*this);
}

TreeBase* RankedTree::plunder() && {
	return new RankedTree(std::move(*this));
}

RankedTree& RankedTree::operator=(const RankedTree& other) {
	if (this == &other) {
		return *this;
	}

	*this = RankedTree(other);

	return *this;
}

RankedTree& RankedTree::operator=(RankedTree&& other) noexcept {
	std::swap(this->tree, other.tree);
	std::swap(this->alphabet, other.alphabet);
	return *this;
}

RankedTree::~RankedTree() noexcept {
	delete tree;
}

const RankedNode& RankedTree::getRoot() const {
	return *tree;
}

RankedNode& RankedTree::getRoot() {
	return *tree;
}

void RankedTree::setTree(RankedNode tree) {
	delete this->tree;
	this->tree = std::move(tree).plunder();
	if(!this->tree->attachTree(this)) {
		delete this->tree;
		throw exception::AlibException("Input symbols not in the alphabet.");
	}
}

bool RankedTree::removeSymbolFromAlphabet(const alphabet::RankedSymbol & symbol) {
	if(this->tree->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	return alphabet.erase(symbol);
}

void RankedTree::operator >>(std::ostream& out) const {
	out << "(RankedTree " << *(this->tree) << ")";
}

int RankedTree::compare(const RankedTree& other) const {
	int res = tree->compare(*other.tree);
	if(res == 0) {
		std::compare<std::set<alphabet::RankedSymbol>> comp;
		res = comp(alphabet, other.alphabet);
	}
	return res;
}

void RankedTree::nicePrint(std::ostream & os) const {
	tree -> nicePrint(os);
}

RankedTree::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string RankedTree::XML_TAG_NAME = "RankedTree";

RankedTree RankedTree::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RankedTree::XML_TAG_NAME);
	std::set<alphabet::RankedSymbol> rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet(input);
	RankedNode * root = TreeFromXMLParser::parseRankedNode(input);
	RankedTree tree(std::move(rankedAlphabet), std::move(*root));
	delete root;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RankedTree::XML_TAG_NAME);
	return tree;
}

void RankedTree::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RankedTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeNode(out, *tree);
	out.emplace_back(RankedTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::RankedTree> rankedTreeParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::RankedTree>(tree::RankedTree::XML_TAG_NAME, tree::RankedTree::parse);
xmlApi<alib::Object>::ParserRegister<tree::RankedTree> rankedTreeParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::RankedTree>(tree::RankedTree::XML_TAG_NAME, tree::RankedTree::parse);

} /* namespace alib */
