/*
 * UnrankedNode.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "UnrankedNode.h"
#include "UnrankedTree.h"
#include "../TreeException.h"
#include "../ranked/RankedNode.h"
#include <sstream>
#include "../../primitive/Unsigned.h"

namespace tree {

UnrankedNode::UnrankedNode ( alphabet::Symbol symbol, std::vector < UnrankedNode * > children ) : symbol ( std::move ( symbol ) ), children ( std::move ( children ) ), parentTree ( NULL ) {
	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
}

UnrankedNode::UnrankedNode ( const UnrankedNode & other ) : symbol ( other.symbol ), parentTree ( NULL ) {
	for ( const auto & element : other.children )
		children.push_back ( element->clone ( ) );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
}

UnrankedNode::UnrankedNode ( UnrankedNode && other ) noexcept : symbol ( std::move ( other.symbol ) ), children ( std::move ( other.children ) ), parentTree ( NULL ) {
	other.children.clear ( );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( NULL );
}

UnrankedNode & UnrankedNode::operator =( const UnrankedNode & other ) {
	if ( this == & other )
		return * this;

	* this = UnrankedNode ( other );

	return * this;
}

UnrankedNode & UnrankedNode::operator =( UnrankedNode && other ) noexcept {
	std::swap ( this->symbol, other.symbol );
	std::swap ( this->children, other.children );
	std::swap ( this->parentTree, other.parentTree ); // this->parentTree is stored within other.parentTree and it is reattached on the next line

	for ( auto & element : this->children )
		element->parent = this;

	this->attachTree ( other.parentTree );

	return * this;
}

UnrankedNode::~UnrankedNode ( ) noexcept {
	for ( auto element : children )
		delete element;

	children.clear ( );
}

const std::vector < const UnrankedNode * > & UnrankedNode::getChildren ( ) const {
	return * reinterpret_cast < const std::vector < const UnrankedNode * > * > ( & children );
}

const std::vector < UnrankedNode * > & UnrankedNode::getChildren ( ) {
	return children;
}

UnrankedNode * UnrankedNode::getParent ( ) {
	return parent;
}

const UnrankedNode * UnrankedNode::getParent ( ) const {
	return parent;
}

void UnrankedNode::setSymbol ( alphabet::Symbol symbol ) {
	this->symbol = std::move ( symbol );

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( this->symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) throw TreeException ( "Symbol is not in the alphabet" );
}

void UnrankedNode::pushBackChild ( const UnrankedNode & node ) {
	UnrankedNode * tmp = node.clone ( );

	tmp->attachTree ( this->parentTree );
	this->children.push_back ( tmp );
}

void UnrankedNode::swap ( UnrankedNode & other ) {
	const UnrankedAlphabet * thisParentTree	 = this->parentTree;
	const UnrankedAlphabet * otherParentTree = other.parentTree;

	UnrankedNode tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachTree ( thisParentTree );
	other.attachTree ( otherParentTree );
}

UnrankedNode * UnrankedNode::clone ( ) const {
	return new UnrankedNode ( * this );
}

UnrankedNode * UnrankedNode::plunder ( ) && {
	return new UnrankedNode ( std::move ( * this ) );
}

RankedNode * UnrankedNode::cloneAsRanked ( ) const {
	std::vector < RankedNode * > clonedChildren;

	for ( const UnrankedNode * child : this->children )
		clonedChildren.push_back ( child->cloneAsRanked ( ) );

	return new RankedNode ( alphabet::RankedSymbol ( this->symbol, primitive::Unsigned ( clonedChildren.size ( ) ) ), std::move ( clonedChildren ) );
}

int UnrankedNode::compare ( const UnrankedNode & other ) const {
	std::compare < alphabet::Symbol > comp;
	int res = comp ( this->symbol, other.symbol );

	if ( res != 0 ) return res;

	int thisSize  = this->children.size ( );
	int otherSize = other.children.size ( );

	if ( thisSize < otherSize ) return -1;

	if ( thisSize > otherSize ) return 1;

	auto thisIter  = this->children.begin ( );
	auto otherIter = other.children.begin ( );

	for ( ; thisIter != this->children.end ( ); ++thisIter, ++otherIter ) {
		int res = ( * thisIter )->compare ( * * otherIter );

		if ( res != 0 ) return res;
	}

	return 0;
}

void UnrankedNode::operator >>( std::ostream & out ) const {
	out << "(UnrankedNode " << " symbol = " << this->symbol << " children = " << this->children << "})";
}

std::ostream & operator <<( std::ostream & out, const UnrankedNode & node ) {
	node >> out;
	return out;
}

bool UnrankedNode::testSymbol ( const alphabet::Symbol & symbol ) const {
	if ( symbol == this->symbol ) return true;

	for ( const auto & child : this->children )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool UnrankedNode::attachTree ( const UnrankedAlphabet * tree ) {
	if ( this->parentTree == tree ) return true;

	this->parentTree = tree;

	for ( const auto & child : this->children )
		if ( !child->attachTree ( tree ) ) return false;

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( this->symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) return false;

	return true;
}

void UnrankedNode::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	alphabet.insert ( this->symbol );

	for ( const auto & child : this->children )
		child->computeMinimalAlphabet ( alphabet );
}

const alphabet::Symbol & UnrankedNode::getSymbol ( ) const {
	return this->symbol;
}

UnrankedNode::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

void UnrankedNode::nicePrint ( std::ostream & os, const std::string & prefix, const bool last ) const {
	os << prefix;

	std::string nextPrefix ( prefix );

	if ( last ) {
		os << "\\-";
		nextPrefix += "  ";
	} else {
		os << "|-";
		nextPrefix += "| ";
	}

	os << ( std::string ) symbol << std::endl;

	for ( unsigned int i = 0; i < children.size ( ); ++i ) {
		os << nextPrefix << "|" << std::endl;
		children[i]->nicePrint ( os, nextPrefix, i == children.size ( ) - 1 );
	}
}

} /* namespace tree */
