/*
 * UnrankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "UnrankedPattern.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedPattern.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace tree {

UnrankedPattern::UnrankedPattern(const RankedPattern& other) : UnrankedPatternAlphabet(other.getSubtreeWildcard().getSymbol()) {
	this->pattern = NULL;
	UnrankedNode* element = other.getRoot().cloneAsUnranked();
	element->computeMinimalAlphabet(alphabet);
	setTree(std::move(*element));
	delete element;
}

UnrankedPattern::UnrankedPattern(alphabet::Symbol subtreeWildcard, std::set<alphabet::Symbol> alphabet, UnrankedNode pattern) : UnrankedPatternAlphabet(std::move(subtreeWildcard)) {
	setAlphabet(std::move(alphabet));
	this->pattern = NULL;
	setTree(std::move(pattern));
}

UnrankedPattern::UnrankedPattern(alphabet::Symbol subtreeWildcard, UnrankedNode pattern) : UnrankedPatternAlphabet(std::move(subtreeWildcard)) {
	pattern.computeMinimalAlphabet(alphabet);
	this->pattern = NULL;
	setTree(std::move(pattern));
}

UnrankedPattern::UnrankedPattern(const UnrankedPattern& other) : UnrankedPatternAlphabet(other), pattern(other.pattern->clone()) {
	this->pattern->attachTree(this);
}

UnrankedPattern::UnrankedPattern(UnrankedPattern&& other) noexcept : UnrankedPatternAlphabet(other), pattern(other.pattern) {
	this->pattern->attachTree(this);
	other.pattern = NULL;
}

TreeBase* UnrankedPattern::clone() const {
	return new UnrankedPattern(*this);
}

TreeBase* UnrankedPattern::plunder() && {
	return new UnrankedPattern(std::move(*this));
}

UnrankedPattern& UnrankedPattern::operator=(const UnrankedPattern& other) {
	if (this == &other) {
		return *this;
	}

	*this = UnrankedPattern(other);

	return *this;
}

UnrankedPattern& UnrankedPattern::operator=(UnrankedPattern&& other) noexcept {
	std::swap(this->pattern, other.pattern);
	std::swap(this->alphabet, other.alphabet);
	std::swap(this->subtreeWildcard, other.subtreeWildcard);
	return *this;
}

UnrankedPattern::~UnrankedPattern() noexcept {
	delete pattern;
}

const UnrankedNode& UnrankedPattern::getRoot() const {
	return *pattern;
}

UnrankedNode& UnrankedPattern::getRoot() {
	return *pattern;
}

void UnrankedPattern::setTree(UnrankedNode pattern) {
	delete this->pattern;
	this->pattern = std::move(pattern).plunder();
	if(!this->pattern->attachTree(this)) {
		delete this->pattern;
		throw exception::AlibException("Input symbols not in the alphabet.");
	}
}

bool UnrankedPattern::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(this->pattern->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	if(this->subtreeWildcard == symbol)
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is subtree wildcard.");

	return alphabet.erase(symbol);
}

void UnrankedPattern::operator >>(std::ostream& out) const {
	out << "(UnrankedPattern " << *(this->pattern) << ")";
}

int UnrankedPattern::compare(const UnrankedPattern& other) const {
	int res = pattern->compare(*other.pattern);
	if(res == 0) {
		std::compare<alphabet::Symbol> comp;
		res = comp(subtreeWildcard, other.subtreeWildcard);
	}
	if(res == 0) {
		std::compare<std::set<alphabet::Symbol>> comp;
		res = comp(alphabet, other.alphabet);
	}
	return res;
}

void UnrankedPattern::nicePrint(std::ostream & os) const {
	pattern -> nicePrint(os);
}

UnrankedPattern::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string UnrankedPattern::XML_TAG_NAME = "UnrankedPattern";

UnrankedPattern UnrankedPattern::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, UnrankedPattern::XML_TAG_NAME);
	alphabet::Symbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol(input);
	std::set<alphabet::Symbol> rankedAlphabet = TreeFromXMLParser::parseAlphabet(input);
	UnrankedNode * root = TreeFromXMLParser::parseUnrankedNode(input);
	UnrankedPattern tree(std::move(subtreeWildcardSymbol), std::move(rankedAlphabet), std::move(*root));
	delete root;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, UnrankedPattern::XML_TAG_NAME);
	return tree;
}

void UnrankedPattern::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(UnrankedPattern::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeSubtreeWildcard(out, subtreeWildcard);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeNode(out, *pattern);
	out.emplace_back(UnrankedPattern::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::UnrankedPattern> unrankedPatternParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::UnrankedPattern>(tree::UnrankedPattern::XML_TAG_NAME, tree::UnrankedPattern::parse);
xmlApi<alib::Object>::ParserRegister<tree::UnrankedPattern> unrankedPatternParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::UnrankedPattern>(tree::UnrankedPattern::XML_TAG_NAME, tree::UnrankedPattern::parse);

} /* namespace alib */
