/*
 * UnrankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "UnrankedTree.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedTree.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace tree {

UnrankedTree::UnrankedTree(const RankedTree& other) {
	this->tree = NULL;
	UnrankedNode* element = other.getRoot().cloneAsUnranked();
	element->computeMinimalAlphabet(alphabet);
	setTree(std::move(*element));
	delete element;
}

UnrankedTree::UnrankedTree(std::set<alphabet::Symbol> alphabet, UnrankedNode tree) {
	setAlphabet(std::move(alphabet));
	this->tree = NULL;
	setTree(std::move(tree));
}

UnrankedTree::UnrankedTree(UnrankedNode tree) {
	tree.computeMinimalAlphabet(alphabet);
	this->tree = NULL;
	setTree(std::move(tree));
}

UnrankedTree::UnrankedTree(const UnrankedTree& other) : tree(other.tree->clone()) {
	alphabet = other.alphabet;
	this->tree->attachTree(this);
}

UnrankedTree::UnrankedTree(UnrankedTree&& other) noexcept : tree(other.tree) {
	alphabet = std::move(other.alphabet);
	this->tree->attachTree(this);
	other.tree = NULL;
}

TreeBase* UnrankedTree::clone() const {
	return new UnrankedTree(*this);
}

TreeBase* UnrankedTree::plunder() && {
	return new UnrankedTree(std::move(*this));
}

UnrankedTree& UnrankedTree::operator=(const UnrankedTree& other) {
	if (this == &other) {
		return *this;
	}

	*this = UnrankedTree(other);

	return *this;
}

UnrankedTree& UnrankedTree::operator=(UnrankedTree&& other) noexcept {
	std::swap(this->tree, other.tree);
	std::swap(this->alphabet, other.alphabet);
	return *this;
}

UnrankedTree::~UnrankedTree() noexcept {
	delete tree;
}

const UnrankedNode& UnrankedTree::getRoot() const {
	return *tree;
}

UnrankedNode& UnrankedTree::getRoot() {
	return *tree;
}

void UnrankedTree::setTree(UnrankedNode tree) {
	delete this->tree;
	this->tree = std::move(tree).plunder();
	if(!this->tree->attachTree(this)) {
		delete this->tree;
		throw exception::AlibException("Input symbols not in the alphabet.");
	}
}

bool UnrankedTree::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(this->tree->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	return alphabet.erase(symbol);
}

void UnrankedTree::operator >>(std::ostream& out) const {
	out << "(UnrankedTree " << *(this->tree) << ")";
}

int UnrankedTree::compare(const UnrankedTree& other) const {
	int res = tree->compare(*other.tree);
	if(res == 0) {
		std::compare<std::set<alphabet::Symbol>> comp;
		res = comp(alphabet, other.alphabet);
	}
	return res;
}

void UnrankedTree::nicePrint(std::ostream & os) const {
	tree -> nicePrint(os);
}

UnrankedTree::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string UnrankedTree::XML_TAG_NAME = "UnrankedTree";

UnrankedTree UnrankedTree::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, UnrankedTree::XML_TAG_NAME);
	std::set<alphabet::Symbol> rankedAlphabet = TreeFromXMLParser::parseAlphabet(input);
	UnrankedNode * root = TreeFromXMLParser::parseUnrankedNode(input);
	UnrankedTree tree(std::move(rankedAlphabet), std::move(*root));
	delete root;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, UnrankedTree::XML_TAG_NAME);
	return tree;
}

void UnrankedTree::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(UnrankedTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeNode(out, *tree);
	out.emplace_back(UnrankedTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::UnrankedTree> unrankedTreeParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::UnrankedTree>(tree::UnrankedTree::XML_TAG_NAME, tree::UnrankedTree::parse);
xmlApi<alib::Object>::ParserRegister<tree::UnrankedTree> unrankedTreeParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::UnrankedTree>(tree::UnrankedTree::XML_TAG_NAME, tree::UnrankedTree::parse);

} /* namespace alib */
