/*
 * UnrankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_TREE_H_
#define UNRANKED_TREE_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "UnrankedNode.h"
#include "../TreeBase.h"
#include "../common/UnrankedAlphabet.h"

namespace tree {

class UnrankedNode;
class RankedTree;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnrankedTree : public TreeBase, public UnrankedAlphabet {
protected:
	UnrankedNode * tree;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual TreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual TreeBase * plunder ( ) &&;

	explicit UnrankedTree ( const RankedTree & other );
	explicit UnrankedTree ( std::set < alphabet::Symbol > alphabet, UnrankedNode tree );
	explicit UnrankedTree ( UnrankedNode tree );

	/**
	 * Copy constructor.
	 * @param other tree to copy
	 */
	UnrankedTree ( const UnrankedTree & other );
	UnrankedTree ( UnrankedTree && other ) noexcept;
	UnrankedTree & operator =( const UnrankedTree & other );
	UnrankedTree & operator =( UnrankedTree && other ) noexcept;
	~UnrankedTree ( ) noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnrankedNode & getRoot ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	UnrankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( UnrankedNode tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const alphabet::Symbol & symbol );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static UnrankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
struct compare < tree::UnrankedTree > {
	int operator ()( const tree::UnrankedTree & first, const tree::UnrankedTree & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* UNRANKED_TREE_H_ */
