#include "AutomatonTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "automaton/FSM/DFA.h"
#include "automaton/FSM/ExtendedNFA.h"
#include "automaton/PDA/SinglePopDPDA.h"
#include "automaton/PDA/DPDA.h"
#include "automaton/PDA/NPDA.h"
#include "automaton/PDA/RealTimeHeightDeterministicNPDA.h"
#include "automaton/TA/NFTA.h"

#include "automaton/AutomatonException.h"

#include "factory/XmlDataFactory.hpp"

#include "alphabet/LabeledSymbol.h"
#include "alphabet/BottomOfTheStackSymbol.h"

#include "regexp/RegExp.h"
#include "regexp/unbounded/UnboundedRegExp.h"
#include "regexp/unbounded/UnboundedRegExpElements.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AutomatonTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( AutomatonTest );

void AutomatonTest::setUp() {
}

void AutomatonTest::tearDown() {
}

void AutomatonTest::testDFAParser() {
	automaton::DFA automaton(automaton::State(1));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));
	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2));
	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), automaton::State(1));

	automaton.addFinalState(automaton::State(3));

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::DFA automaton2 = alib::XmlDataFactory::fromTokens<automaton::DFA>(tokens2);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::SinglePopDPDATransitions() {
	automaton::SinglePopDPDA automaton(automaton::State(1), alphabet::symbolFrom("S"));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addStackSymbol(alphabet::symbolFrom("S"));
	automaton.addStackSymbol(alphabet::symbolFrom("X"));
	automaton.addStackSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(automaton::State(3), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(3), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("Y"), automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), /* eps, */ alphabet::symbolFrom("Y"), automaton::State(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(2), /* eps, */ alphabet::symbolFrom("X"), automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addFinalState(automaton::State(3));

}

void AutomatonTest::DPDATransitions() {
	automaton::DPDA automaton(automaton::State(1), alphabet::symbolFrom("S"));

	automaton.addState(automaton::State(1));
	automaton.addState(automaton::State(2));
	automaton.addState(automaton::State(3));
	automaton.addState(automaton::State(4));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addStackSymbol(alphabet::symbolFrom("S"));
	automaton.addStackSymbol(alphabet::symbolFrom("X"));
	automaton.addStackSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(automaton::State(3), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(3), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("Y") }, automaton::State(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(1), /* eps, */ { alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(2), /* eps, */ { alphabet::symbolFrom("X") }, automaton::State(1), { alphabet::symbolFrom("X") } ),
		exception::AlibException
	);

	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("S") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT(!automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } ));
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } ),
		exception::AlibException
	);
	automaton.addTransition(automaton::State(4), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, automaton::State(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(automaton::State(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, automaton::State(2), { alphabet::symbolFrom("Y") } ),
		exception::AlibException
	);

	automaton.addFinalState(automaton::State(3));
}

void AutomatonTest::testExtendedNFAAlphabet() {
	automaton::ExtendedNFA automaton (automaton::State(0));
	automaton::State s0(0);
	automaton::State s1(1);
	automaton.addState(s0);
	automaton.addState(s1);

	automaton.setInputAlphabet({alphabet::symbolFrom('a'), alphabet::symbolFrom('b') });

	regexp::UnboundedRegExpConcatenation goodConcat1, goodConcat2;
	regexp::UnboundedRegExpConcatenation badConcat;

	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('a')));
	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('d')));

	CPPUNIT_ASSERT_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(badConcat)}, s1), automaton::AutomatonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat1)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat2)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpEpsilon())}, s1));
}

void AutomatonTest::testNPDATransitions() {
	automaton::NPDA automaton(automaton::State(0), alphabet::symbolFrom("T"));

	automaton.setStates({automaton::State(0), automaton::State(1), automaton::State(2), automaton::State(3), automaton::State(4)});
	automaton.setInputAlphabet({alphabet::symbolFrom("a2"), alphabet::symbolFrom("a1"), alphabet::symbolFrom("a0")});
	automaton.setStackAlphabet({alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});

	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(0), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(0), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(0), {});

	automaton.addTransition(automaton::State(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("R"), alphabet::symbolFrom("T")});

	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, automaton::State(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(1), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, automaton::State(1), {alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(1), {});
	automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, automaton::State(2), {});

	automaton.addTransition(automaton::State(2), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("R")});

	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, automaton::State(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, automaton::State(3), {alphabet::symbolFrom("T")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, automaton::State(3), {alphabet::symbolFrom("R")});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, automaton::State(3), {});
	automaton.addTransition(automaton::State(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, automaton::State(4), {});

	automaton.setFinalStates({automaton::State(4)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testRHPDATransitions() {
	automaton::RealTimeHeightDeterministicNPDA automaton{alphabet::Symbol(alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK)};
	automaton.setStates({automaton::State(1), automaton::State(2), automaton::State(3)});
	automaton.setInitialStates({automaton::State(1)});

	automaton.setInputAlphabet({alphabet::symbolFrom("a"), alphabet::symbolFrom("b")});
	automaton.setStackAlphabet({alphabet::Symbol(alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK), alphabet::symbolFrom("X"), alphabet::symbolFrom("Y")});

	automaton.addCallTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2), alphabet::symbolFrom("X"));
	automaton.addCallTransition(automaton::State(2), automaton::State(3), alphabet::symbolFrom("X"));
	automaton.addReturnTransition(automaton::State(3), alphabet::symbolFrom("Y"), automaton::State(1));

	automaton.setFinalStates({automaton::State(3)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testNFTAParser() {
	alphabet::RankedSymbol a ('a', 2);
	alphabet::RankedSymbol b ('b', 1);
	alphabet::RankedSymbol c ('c', 0);

	automaton::State q2 ("q2");
	automaton::State q1 ("q1");
	automaton::State q0 ("q0");

	automaton::NFTA automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<automaton::State> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<automaton::State> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<automaton::State> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << tmp << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::NFTA automaton2 = alib::XmlDataFactory::fromTokens<automaton::NFTA>(tokens2);

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::testNFTATransitions() {
	alphabet::RankedSymbol a ('a', 2);
	alphabet::RankedSymbol b ('b', 1);
	alphabet::RankedSymbol c ('c', 0);
	alphabet::RankedSymbol d ('d', 0);

	automaton::State q2 ("q2");
	automaton::State q1 ("q1");
	automaton::State q0 ("q0");

	automaton::NFTA automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<automaton::State> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<automaton::State> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<automaton::State> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT_THROW(automaton.removeInputSymbol(a), exception::AlibException);
	CPPUNIT_ASSERT_NO_THROW(automaton.removeInputSymbol(d));
	CPPUNIT_ASSERT_NO_THROW(automaton.removeTransition(b, states1, q2));
	CPPUNIT_ASSERT_THROW(automaton.addTransition(b, states1, q1), exception::AlibException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(a, states2, q1), exception::AlibException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(d, states3, q0), exception::AlibException);
}
